--[[
Snowflake - a library to provide a dropdown menu interface for professions

Copyright (C) 2006-2008  Prat Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:
	
Free Software Foundation, I.,
51 Franklin Street, Fifth Floor,
Boston, MA  02110-1301, USA.
--]]

local MAJOR, MINOR = 'Snowflake-2.0', 1
local Snowflake = LibStub:NewLibrary( MAJOR, MINOR)
if not Snowflake then return end -- already loaded and no upgrade necessary

Snowflake.name = MAJOR
setmetatable( Snowflake, { __tostring = function() return MAJOR end })

local AceGUI = LibStub("AceGUI-3.0")
local AceConsole = LibStub( 'AceConsole-3.0')

local function SetTooltip( btn)
	local link = btn.userdata.link
	if link then
		GameTooltip_SetDefaultAnchor( GameTooltip, btn.frame);
		GameTooltip:SetHyperlink( link)
	end
end

local function Call( func, par, ...)
	if func then
		local temp = {}
		for _,v in ipairs(par) do
			table.insert(temp, v)
		end
		for i = 1,select( '#', ...) do 
			table.insert(temp, (select( i, ...)))
		end
		local res, msg
		local this = par[1]
		if type(func) == 'function' then
			res, msg = pcall( func, unpack(temp))
		elseif type(this) == 'table' and type(this[func]) == 'function' then
			res, msg = pcall( this[func], unpack(temp))
		else
			msg = 'not a function'
		end
		if res then 
			return msg
		else
			AceConsole:Print( 'Snowflake: Cannot call method:', func, msg)
		end
	end
end

local function Button_OnEnter( btn)
--	AceConsole:Print( 'Button_OnEnter', btn.parent and btn.parent.frame:GetName(), btn.frame:GetName())
	local tbl = btn.userdata
	if tbl.main then
		tbl.main:ResetTimer()
	end
	btn:SetSelect( true)
	if btn.parent then
		btn.parent:SetToken( tbl.token)
		if tbl.hasArrow then
			btn.parent:Neighbour( tbl.func, tbl.args)
		else
			btn.parent:CloseNeighbour()
		end
	else
		AceConsole:Print( 'Snowflake: Button_OnEnter', btn.frame:GetName(), 'oh no, missing parent group')
	end
	SetTooltip( btn)
end

local function Button_OnLeave( btn)
	GameTooltip:Hide()
end

local function Button_OnClick( btn)
	local tbl = btn.userdata
	if not (tbl.disabled or tbl.hasArrow) then
		Call( tbl.func, tbl.args)
	end
	if tbl.main then
		if tbl.doClose then
			tbl.main:Fire( 'OnClose')
		else
			tbl.main:Refresh()
		end
	end
end

local function Button_OnEnterPressed( btn, name, txt)
	local tbl = btn.userdata
	if not tbl.disabled then
		Call( tbl.func, tbl.args, txt)
	end
	if tbl.main then
		if tbl.doClose then
			tbl.main:Fire( 'OnClose')
		else
			tbl.main:Refresh()
		end
	end
end

local function Button_OnMouseDown( btn)
	local tbl = btn.userdata
	if tbl.main then -- and (tbl.disabled or tbl.hasArrow) 
		tbl.main:Fire( 'OnMouseDown')
	end
end

local function Button_OnMouseUp( btn)
	local tbl = btn.userdata
	if tbl.main then -- and (tbl.disabled or tbl.hasArrow) 
		tbl.main:Fire( 'OnMouseUp')
	end
end

local function Group_OnAddLine( grp, name, param)
	local info = param or { disabled = true }
	info.token = info.token or info.text or grp:NextSpaceName()
	local newType = 'SnowflakeButton'
	if info.editbox then
		newType = 'SnowflakeEditBox'
	end
	local btn = grp:GetOldChild( info.token)
	if not btn then
		btn = AceGUI:Create( newType)
	elseif btn.type ~= newType then
		AceGUI:Release( btn)
		btn = AceGUI:Create( newType)
	end
	btn.userdata.main = grp.userdata.main
	btn.width = 'fill'
	btn:SetInfo( info)
	btn:SetCallback( 'OnEnter', Button_OnEnter)
	btn:SetCallback( 'OnLeave', Button_OnLeave)
	btn:SetCallback( 'OnClick', Button_OnClick)
	btn:SetCallback( 'OnMouseDown', Button_OnMouseDown)
	btn:SetCallback( 'OnMouseUp',   Button_OnMouseUp)
	btn:SetCallback( 'OnEnterPressed', Button_OnEnterPressed)
	grp:AddChild( btn)
end

local function Dock( grp, btn, isBottom, isLeft, offset)
	grp:ClearAllPoints()
	if isBottom then
		if isLeft then
			grp:SetPoint( 'BOTTOMRIGHT', grp.parent.frame, 'BOTTOMLEFT',  0, offset or 0)
		else
			grp:SetPoint( 'BOTTOMLEFT', grp.parent.frame, 'BOTTOMRIGHT', 0, offset or 0)
		end
	else
		if isLeft then
			grp:SetPoint( 'TOPRIGHT', grp.parent.frame, 'TOPLEFT',  0, -offset or 0)
		else
			grp:SetPoint( 'TOPLEFT', grp.parent.frame, 'TOPRIGHT', 0, -offset or 0)
		end
	end
end

local function Group_OnDock( grp, name, func, args)
--	AceConsole:Print( 'Group_OnDock', grp.frame:GetName(), grp.neighbour.frame:GetName())
	local neighbour = grp.neighbour
	neighbour:SetFunction( func, args)
	neighbour:Refresh()
	local btn, offset = grp:GetTokenChild()
	Dock( neighbour, btn, false, false, offset)
	local frame = neighbour.frame
	local isBottom = (frame:GetTop() or 0) > GetScreenHeight()
	local isLeft   = (frame:GetRight() or 0) > GetScreenWidth()
	if isBottom or isLeft then
		Dock( neighbour, btn, isBottom, isLeft, offset)
	end
	local top    = frame:GetTop() or 0
	local bottom = frame:GetBottom() or 0
	if top > GetScreenHeight() then
		local point, relativeTo, relativePoint, x, y = frame:GetPoint(1)
		frame:ClearAllPoints()
		frame:SetPoint( point, relativeTo, relativePoint, x or 0, (y or 0) + GetScreenHeight() - top)
	elseif bottom < 0 then
		local point, relativeTo, relativePoint, x, y = frame:GetPoint(1)
		frame:ClearAllPoints()
		frame:SetPoint(point, relativeTo, relativePoint, x or 0, (y or 0) - bottom)
	end
end

local function Group_OnNeighbour( grp, name, func, args)
--	AceConsole:Print( 'Group_OnNeighbour', grp.frame:GetName())
	local neighbour = grp.neighbour
	if not neighbour then
		neighbour = AceGUI:Create( 'SnowflakeGroup')
		neighbour:SetLayout( 'ListZoom')
		neighbour:SetCallback( 'OnAddLine',   Group_OnAddLine)
		neighbour:SetCallback( 'OnNeighbour', Group_OnNeighbour)
		neighbour:SetCallback( 'OnDock',      Group_OnDock)
		grp:SetNeighbour( neighbour)
	end
	grp:DockNeighbour( func, args)
	neighbour.frame:Show()
end

function Snowflake:CreateMain()
	local grp = AceGUI:Create( 'SnowflakeGroup')
	grp:SetLayout( 'ListZoom')
	grp:SetCallback( 'OnAddLine',   Group_OnAddLine)
	grp:SetCallback( 'OnNeighbour', Group_OnNeighbour)
	grp:SetCallback( 'OnDock',      Group_OnDock)
	grp.userdata.main = grp
	return grp
end

AceGUI:RegisterLayout( 'ListZoom', function( content, children)
	local height = 0
	local width = 0
	for i,child in ipairs(children) do
		local frame = child.frame
		frame:ClearAllPoints()
		frame:Show()
		if i == 1 then
			frame:SetPoint( 'TOPLEFT', content, 'TOPLEFT', 0, 0)
		else
			frame:SetPoint( 'TOPLEFT', children[i-1].frame, 'BOTTOMLEFT', 0, 0)
		end
		if child.DoLayout then
			child:DoLayout()
		end
		if child.width == 'fill' then
			if child.GetMinWidth then
				width = math.max(width, child:GetMinWidth())
			end
			frame:SetPoint( 'RIGHT', content, 'RIGHT', 0, 0)
		end
		height = height + (frame.height or frame:GetHeight() or 0)
	end
	content.obj:LayoutFinished( width, height )
end)
